package UDP::Corpus;

use Moose;
use UDP::Sentence;
use UDP::Node;
use Graph;
use Graph::Directed;
use Graph::ChuLiuEdmonds;

has 'sentences' => ( is => 'rw', isa => 'ArrayRef', default => sub { [] } );
has 'logprob'   => ( is => 'rw', isa => 'Num' );
has 'counts'    => ( is => 'rw', isa => 'HashRef' );
has 'edge_cnt'  => ( is => 'rw', isa => 'HashRef', default => sub {my %h = (); return \%h;} );


sub get_sentences {
    my ($self) = @_;
    return @{$self->sentences};
}

sub delete_counts {
    my ($self) = @_;
    my %empty = ();
    $self->counts(\%empty);
}

sub update_counts {
    my ($self, $diff_counts) = @_;
    foreach my $key (keys %$diff_counts) {
        $self->counts->{$key} += $diff_counts->{$key};
        delete $self->counts->{$key} if !$self->counts->{$key};
    }
}


sub load_from_conll {
    my ($self, $form_column, $tag_column) = @_;
    my $nodes = [];
    my $w = 0;
    while (<>) {
        chomp;
        if (/^\s*$/) {
            my $sentence = UDP::Sentence->new({nodes => $nodes});
            map {$_->sentence($sentence)} @$nodes;
            push @{$self->sentences}, $sentence;
            $nodes = [];
            $w = 0;
        }
        else {
            my @items = split /\t/;
            $$nodes[$w] = UDP::Node->new({tag => $items[$tag_column], form => $items[$form_column], ord => ($w + 1), leftmost => ($w + 1), rightmost => ($w + 1)});
            $w++;
        }
    }
}

sub print_to_conll {
    my ($self) = @_;
    foreach my $sentence ($self->get_sentences) {
        foreach my $node ($sentence->get_nodes) {
            print $node->ord."\t".$node->form."\t_\t_\t".$node->tag."\t_\t".$node->parent->ord."\t_\t_\t_\n";
        }
        print "\n";
    }
}

sub initialize {
    my ($self) = @_;
    foreach my $sentence ($self->get_sentences) {
        $sentence->initialize();
    }
}

sub compute_logprob {
    my ($self, $model) = @_;
    my $diff_logprob = 0;
    my %diff_counts = ();
    $self->delete_counts();
    foreach my $sentence ($self->get_sentences) {
        foreach my $node ($sentence->get_nodes) {
            $model->compute_counts_and_logprob('ins', $node, \%diff_counts, \$diff_logprob, $self->counts);
        }
    }
    $self->logprob($diff_logprob);
    $self->delete_counts();
    $self->update_counts(\%diff_counts);
}

sub gibbs_sampling {
    my ($self, $model, $iterations) = @_;
    foreach my $iter (1 .. $iterations) {
        foreach my $sentence ($self->get_sentences) {
            my @shuffled_nodes = List::Util::shuffle ($sentence->get_nodes);
            foreach my $node1 (@shuffled_nodes) {
                next if $node1->parent->tag eq '<root>';
                my $node2 = $node1->parent;
                my $grandparent = $node2->parent;
                my @active = sort {$a->ord <=> $b->ord} ($node1->get_children, $node2->get_children, $node2);
                ($node1, $node2) = ($node2, $node1) if $node1->ord > $node2->ord;
                my ($pos1, $pos2);
                foreach my $i (0 .. $#active) {
                    $pos1 = $i if $active[$i] eq $node1;
                    $pos2 = $i if $active[$i] eq $node2;
                }
                my @gov;
                my @dep;
                my @gov_children;
                my @dep_children;
                # a) node2 will be the new head
                foreach my $left_end (0 .. $pos1) {
                    foreach my $right_end ($pos1 .. ($pos2 - 1)) {
                       push @gov, $node2;
                       my @ch_gov = @active[0 .. ($left_end - 1), $pos1, ($right_end + 1) .. ($pos2 - 1), ($pos2 + 1) .. $#active];
                       push @gov_children, \@ch_gov;
                       push @dep, $node1;
                       my @ch_dep = @active[$left_end .. ($pos1 - 1), ($pos1 + 1) .. $right_end]; 
                       push @dep_children, \@ch_dep;
                    }
                }
                # b) node1 will be the new head
                foreach my $left_end (($pos1 + 1) .. $pos2) {
                    foreach my $right_end ($pos2 .. $#active) {
                       push @gov, $node1;
                       my @ch_gov = @active[0 .. ($pos1 - 1), ($pos1 + 1) .. ($left_end - 1), $pos2, ($right_end + 1) .. $#active];
                       push @gov_children, \@ch_gov;
                       push @dep, $node2;
                       my @ch_dep = @active[$left_end .. ($pos2 - 1), ($pos2 + 1) .. $right_end];
                       push @dep_children, \@ch_dep;
                    }
                }
                # compute scores for individual candidates
                my @diff_count;
                my @new_logprob;
                my @weight;
                my $sum_weight = 0;
                my $current_logprob = $self->logprob;
                my %current_diff = ();
                foreach my $i (0 .. $#gov) {

                    my $different_gov = $dep[$i]->parent ne $gov[$i] ? 1 : 0;

                    # 1) go through all the new children of the new dependent, and if their parent is changed, recompute the logprob
                    foreach my $ch (@{$dep_children[$i]}) {
                        if ($ch->parent ne $dep[$i]) {
                            $model->compute_counts_and_logprob('del', $ch, \%current_diff, \$current_logprob, $self->counts);
                            $ch->parent($dep[$i]);
                            $model->compute_counts_and_logprob('ins', $ch, \%current_diff, \$current_logprob, $self->counts);
                        }
                    }
                    
                    # 2) go through all the new children of the new governor (except of the previous governor), and if their parent is changed, recompute the logprob
                    foreach my $ch (@{$gov_children[$i]}) {
                        if ($ch->parent ne $gov[$i] && $ch ne $dep[$i]) {
                            $model->compute_counts_and_logprob('del', $ch, \%current_diff, \$current_logprob, $self->counts);
                            $ch->parent($gov[$i]);
                            $model->compute_counts_and_logprob('ins', $ch, \%current_diff, \$current_logprob, $self->counts);
                        }
                    }

                    # 3) subtract counts of the previous dependent and governor
                    $model->compute_counts_and_logprob('del', $dep[$i], \%current_diff, \$current_logprob, $self->counts);
                    $model->compute_counts_and_logprob('del', $gov[$i], \%current_diff, \$current_logprob, $self->counts);
                    
                    # 4) add counts of the new dependent and governor
                    $dep[$i]->set_children($dep_children[$i]);
                    $dep[$i]->parent($gov[$i]);
                    $model->compute_counts_and_logprob('ins', $dep[$i], \%current_diff, \$current_logprob, $self->counts);
                    $gov[$i]->set_children($gov_children[$i]);
                    $gov[$i]->parent($grandparent) if $different_gov;
                    $model->compute_counts_and_logprob('ins', $gov[$i], \%current_diff, \$current_logprob, $self->counts);

                    # recompute the counts and logprob for the grandparent
                    if ($different_gov) {
                        my @gr_children = map {$_ eq $dep[$i] ? $gov[$i] : $_} $grandparent->get_children;
                        if ($grandparent->tag ne '<root>') {
                            $model->compute_counts_and_logprob('del', $grandparent, \%current_diff, \$current_logprob, $self->counts);
                        }
                        $grandparent->set_children(\@gr_children);
                        if ($grandparent->tag ne '<root>') {
                            $model->compute_counts_and_logprob('ins', $grandparent, \%current_diff, \$current_logprob, $self->counts);
                        }
                    }

                    # add the parameters of this candidate to the lists
                    my %d = %current_diff;
                    push @diff_count, \%d;
                    push @new_logprob, $current_logprob;
                    push @weight, exp($current_logprob - $self->logprob);
                    $sum_weight += $weight[$#weight];
                }
                # choose one candidate by weighted coin flip
                my $random_value = rand($sum_weight);
                my $current_value = 0;
                foreach my $i (0 .. $#weight) {
                    $current_value += $weight[$i];
                    if ($current_value >= $random_value) {
                        # update logprob
                        $self->logprob($new_logprob[$i]);
                        # update counts
                        $self->update_counts($diff_count[$i]);
                        # update structure
                        $dep[$i]->set_children($dep_children[$i]);
                        map {$_->parent($dep[$i])} @{$dep_children[$i]};
                        $gov[$i]->set_children($gov_children[$i]);
                        map {$_->parent($gov[$i])} @{$gov_children[$i]};
                        $gov[$i]->parent($grandparent);
                        my @gr_children = map {$_ eq $dep[$i] ? $gov[$i] : $_} $grandparent->get_children;
                        $grandparent->set_children(\@gr_children);
                        last;
                    }
                }
            }
            if ($iter > 5 && rand() < 0.01) {
                foreach my $sentence ($self->get_sentences) {
                    foreach my $node ($sentence->get_nodes) {
                        $self->edge_cnt->{$node}{$node->parent}++;
                    }
                }
            }
        }
        print STDERR "Iteration $iter, logprob = ".$self->logprob."\n";
    }
}

sub parse_mst {
    my ($self, $function) = @_;
    foreach my $sentence ($self->get_sentences) {
        my @nodes = ($sentence->root, $sentence->get_nodes);
        my @vertices = (0 .. $#nodes);
        my $graph = Graph::Directed->new( vertices => \@vertices );
        foreach my $w ( 1 .. $#nodes ) {
            foreach my $g ( 0 .. $#nodes ) {
                next if $w == $g;
                my $edge_count = $self->edge_cnt->{$nodes[$w]}{$nodes[$g]} || 0;
                my $edge_weight;
                if ($function eq 'log') {
                    $edge_weight = $edge_count ? log($edge_count) : -1000000; # logarithm of 0
                }
                else {
                    $edge_weight = $edge_count;
                }
                $graph->add_weighted_edge($g, $w, -$edge_weight); # negative values because of minimum spanning tree algorithm
            }
        }
        my $mst = $graph->MST_ChuLiuEdmonds($graph);
        foreach my $e ($mst->edges) {
            $nodes[$e->[1]]->parent($nodes[$e->[0]]);
        }
    }
}


1;

