#!/usr/bin/env perl

use strict;
use warnings;

use Getopt::Long;
use FindBin qw($Bin);

use lib "$Bin/lib";

use UDP::Corpus;
use UDP::Model;
use UDP::Model::EdgeDir;
use UDP::Model::Distance;
use UDP::Model::Subtree;
use UDP::Model::Fertility;
use UDP::Model::LexicalEdge;
my $TAG_COLUMN = 4;
my $FORM_COLUMN = 1;
my $CONFIG_FILE = "";
my $REDUCIBILITY = "";
my $ITERATIONS = 20;
my $PROJECTIVE = 1;

GetOptions(
    "tag-column=i"   => \$TAG_COLUMN,
    "form-column=i"  => \$FORM_COLUMN,
    "config=s"       => \$CONFIG_FILE,
    "reducibility=s" => \$REDUCIBILITY,
    "iterations=i"   => \$ITERATIONS,
    "projective=i"   => \$PROJECTIVE,
);

# print usage, if not reducibility nor config have been specified
if (!$CONFIG_FILE && !$REDUCIBILITY) {
    print STDERR "USAGE:

./parse.pl --reducibility file_with_reducibility scores < input_conll_file > output_conll_file

";
exit;
}

# set model parameters
my %parameter;
if ($CONFIG_FILE) {
    # read config file
    open (CONFIG, "<:utf8", $CONFIG_FILE) or die "Cannot find config file '$CONFIG_FILE'.";
    print STDERR "Reading configuration file '$CONFIG_FILE'\n";
    while (<CONFIG>) {
        chomp;
        my @items = split /\s/;
        next if $items[0] =~ /^\s*#/;
        my $model_name = shift @items;
        foreach my $item (@items) {
            if ($item =~ /^(.+)=(.+)$/) {
                $parameter{$model_name}{$1} = $2;
            }
        }
    }
}
else {
    # or set default parameters
    print STDERR "Configuration file has not been specified. Default parameters will be used.\n";
    $parameter{'edge'}{'alpha'} = 1;
    $parameter{'distance'}{'alpha'} = 1.5;
    $parameter{'subtree'}{'alpha'} = 1;
    $parameter{'subtree'}{'default_score'} = 0.03;
    $parameter{'subtree'}{'score_file'} = $REDUCIBILITY;
    $parameter{'fertility'}{'alpha'} = 0.01;
    $parameter{'fertility'}{'freq_contribution'} = 'linear';
    $parameter{'mst'}{'type'} = 'linear';
}

print STDERR "\nUsed model:\n";

my @submodels;
if ($parameter{'edge'}) {
    my $model = UDP::Model::EdgeDir->new($parameter{'edge'});
    print STDERR "Edge model: " . (join " ", map {"$_=$parameter{'edge'}{$_}"} keys %{$parameter{'edge'}}) . "\n";
    push @submodels, $model;
}
if ($parameter{'distance'}) {
    my $model = UDP::Model::Distance->new($parameter{'distance'});
    print STDERR "Distance model: " . (join " ", map {"$_=$parameter{'distance'}{$_}"} keys %{$parameter{'distance'}}) . "\n";
    push @submodels, $model;
}
if ($parameter{'subtree'}) {
    my $model = UDP::Model::Subtree->new($parameter{'subtree'});
    print STDERR "Subtree model: " . (join " ", map {"$_=$parameter{'subtree'}{$_}"} keys %{$parameter{'subtree'}}) . "\n";
    push @submodels, $model;
}
if ($parameter{'fertility'}) {
    my $model = UDP::Model::Fertility->new($parameter{'fertility'});
    print STDERR "Fertility model: " . (join " ", map {"$_=$parameter{'fertility'}{$_}"} keys %{$parameter{'fertility'}}) . "\n";
    push @submodels, $model;
}
if ($parameter{'lexical_edge'}) {
    my $model = UDP::Model::LexicalEdge->new($parameter{'lexical_edge'});
    print STDERR "Lexical edge model: " . (join " ", map {"$_=$parameter{'lexical_edge'}{$_}"} keys %{$parameter{'lexical_edge'}}) . "\n";
    push @submodels, $model;
}

my $corpus = UDP::Corpus->new({use_aux => 0});
print STDERR "Reading CoNLL...\n";
$corpus->load_from_conll($FORM_COLUMN, $TAG_COLUMN);

my $model = UDP::Model->new({submodels => \@submodels});

print STDERR "Initializing...\n";
$corpus->initialize();
map {$_->initialize($corpus)} $model->get_submodels;


$corpus->compute_logprob($model);
print STDERR "Logprob: " . $corpus->logprob . "\n";
if ($PROJECTIVE) {
    $corpus->gibbs_sampling($model, $ITERATIONS);
}
else {
    $corpus->nonprojective_gibbs_sampling($model, $ITERATIONS);
}
$corpus->compute_logprob($model);
print STDERR "Logprob: " . $corpus->logprob . "\n";

if ($parameter{'mst'}) {
    print STDERR "\nComputing maximum spanning trees...";
    $corpus->parse_mst($parameter{'mst'}{'type'});
}

print STDERR "Printing...\n";
$corpus->print_to_conll();

